@@ -5,12 +5,13 @@
 #' treatment group. \code{hbal} automatically expands the covariate space to include
 #' higher order terms and uses cross-validation to select variable penalties for the 
 #' balancing conditions.
-#' @usage hbal(data, Treat, X, Y, base.weights = NULL, coefs = NULL ,
-#'      max.iterations = 200, cv = TRUE, folds = 4, 
-#'      expand.degree = 3, ds = FALSE, alpha = NULL, 
-#'      group.alpha=NULL, constraint.tolerance = 1e-3, 
-#'      print.level = -1, grouping = NULL, 
-#'      shuffle.treat=TRUE, exclude=NULL, seed=NULL)
+#' @usage hbal(data, Treat, X, Y = NULL, w = NULL, 
+#'      X.expand = NULL, X.keep = NULL, expand.degree = 1,
+#'      coefs = NULL, max.iterations = 200, cv = NULL, folds = 4,
+#'      ds = FALSE, group.exact = NULL, group.alpha = NULL,
+#'      term.alpha = NULL, constraint.tolerance = 1e-3, print.level = 0,
+#'      grouping = NULL, group.labs = NULL, linear.exact = TRUE, shuffle.treat = TRUE,
+#'      exclude = NULL,force = FALSE, seed = 94035)
 #' @param data                 a dataframe that contains the treatment, outcome, and covariates.   
 #' @param Treat                a character string of the treatment variable.
 #' @param X                    a character vector of covariate names to balance on.
@@ -31,6 +32,7 @@
 #' @param print.level          details of printed output.
 #' @param grouping             different groupings of the covariates. Must be specified if expand is \code{FALSE}.
 #' @param group.labs           labels for user-supplied groups
+#' @param linear.exact		   seek exact balance on the level terms
 #' @param shuffle.treat        whether to use cross-validation on the treated units. Default is \code{TRUE}.
 #' @param exclude              list of covariate name pairs or triplets to be excluded.
 #' @param force                binary indicator of whether to expand covariates when there are too many
@@ -48,6 +50,10 @@
 #' \item{Y}{vector of outcome}
 #' @author Yiqing Xu, Eddie Yang
 #' @importFrom stats var
+#' @importFrom stats setNames
+#' @importFrom stats complete.cases
+#' @importFrom stats weighted.mean
+#' @importFrom stats density
 #' @importFrom nloptr nloptr
 #' @useDynLib hbal, .registration = TRUE
 #' @references Xu, Y., & Yang, E. (2022). Hierarchically Regularized Entropy Balancing. Political Analysis, 1-8. doi:10.1017/pan.2022.12
@@ -94,7 +100,7 @@ hbal <- function(
 	expand.degree = 1,
 	coefs = NULL,
 	max.iterations = 200,
-	cv = FALSE,
+	cv = NULL,
 	folds = 4,
 	ds = FALSE,
 	group.exact = NULL,
@@ -104,10 +110,11 @@ hbal <- function(
 	print.level = 0,
 	grouping = NULL,
 	group.labs = NULL,
+	linear.exact = TRUE,
 	shuffle.treat = TRUE,
 	exclude = NULL,
 	force = FALSE,
-	seed = NULL
+	seed = 94035
 	){
 
 	# ntreated: number of treated units
@@ -133,9 +140,28 @@ hbal <- function(
 	mcall <- match.call()
 	if(!is.null(seed)) set.seed(seed)
 
+	
+	#renames the covariates
+	new_names <- paste0("X", seq_along(X))
+	colnames(data)[colnames(data) %in% X] <- new_names
+	if (is.null(X.expand) == FALSE)
+	{
+	  mapping <- setNames(new_names, X)
+	  mapped_list <- mapping[match(X.expand, names(mapping))]
+	  X.expand <- unname(mapped_list)
+	}
+	if (is.null(X.keep) == FALSE)
+	{
+	  mapping <- setNames(new_names, X)
+	  mapped_list <- mapping[match(X.keep, names(mapping))]
+	  X.keep <- unname(mapped_list)
+	}
+	old_names = X
+	X = new_names
+	
 	# covariates for expansion
-	if (expand.degree >= 0 && is.null(X.expand) == TRUE) {
-		if (print.level > 0) message("All variables will be serially expanded")
+	if (expand.degree > 1 && is.null(X.expand) == TRUE) {
+		if (print.level >= 1) message("All variables will be serially expanded")
 		if ((expand.degree == 2 & length(X)>30) | (expand.degree == 3 & length(X)>10)) {
 			if (force == TRUE) {
 				message("Too many variables to expand. This can take up significant memory.\nPlease consider including fewer covariates or using the \`X.expand\` option\n")
@@ -146,7 +172,7 @@ hbal <- function(
 		X.expand <- X
 	}
 	if (expand.degree == 1 && is.null(X.expand) == FALSE) {
-		if (print.level >= 0) {
+		if (print.level >= 1) {
 			message("\"X.expand\" is ignored because expand.degree = 1; no serial expansion will be performed")}
 		X.expand <- NULL
 	}
@@ -167,7 +193,7 @@ hbal <- function(
 	# listwise deletion
 	valid_rows <- complete.cases(data[, c(Treat, X.all, Y, w)])
 	if (sum(valid_rows) < nrow(data)) {
-		if (print.level >= 0) {message("Some rows are dropped because they contain missing/NA/infinite values\n")}
+		if (print.level >= 1) {message("Some rows are dropped because they contain missing/NA/infinite values\n")}
 		data <- data[valid_rows,]		
 	}
 
@@ -204,7 +230,7 @@ hbal <- function(
 	# Variable positions: X.levelonly.pos, X.expand.pos, X.keep.pos
 
 	# check X variation
-	X.novar <- X.all[which(apply(data[,X.all], 2, sd) == 0)] # controls of no variations
+	X.novar <- X.all[which(apply(data[, X.all, drop = FALSE], 2, sd) == 0)] # controls of no variations
 	if (length(X.novar) > 0 && print.level >= 0) {
 		message(paste("The following variable(s) have no variations and are automatically dropped:", paste0(X.novar, collapse = ", "),"\n"))
 		X.all <- setdiff(X.all, X.novar)
@@ -258,9 +284,9 @@ hbal <- function(
 	# series expansion of the covariates
 	if (expand.degree > 1) {
 		expand <- covarExpand(X.sav[, X.expand.pos, drop = FALSE], 
-			exp.degree = expand.degree, treatment = Treatment, exclude = exclude)
-		X.tmp <- expand$mat
-		grouping <- expand$grouping
+			exp.degree = expand.degree, treatment = Treatment, exclude = exclude) 
+		X.tmp <- expand$mat # expanded covariates in matrix form
+		grouping <- expand$grouping # grouping of the expanded covariates
 		if (expand.degree == 2) {group.labs <- c("linear", "squared", "two-way")} # three of them
 		if (expand.degree == 3) {group.labs <- c("linear", "two-way", "squared", "three-way", "squared*linear", "cubic")} # six of them
 		names(grouping) <- group.labs[1:length(grouping)]
@@ -324,18 +350,12 @@ hbal <- function(
 			grouping[1] <- grouping[1] + (length(X.pos) - length(X.select.pos))			
 		}
 		# update X and X.sav
-		X <- X[, X.pos]		
-		X.sav <- X.sav[, X.pos] 		
-	}
-
-	
-	if (length(grouping)==1){
-		cv <- FALSE
-		if (print.level >= 0) message("length(grouping) = 1, setting \'cv = FALSE\'")
+		X <- X[, X.pos, drop = FALSE]		
+		X.sav <- X.sav[, X.pos, drop = FALSE] 		
 	}
 
 	full.c <- cbind(rep(1,ncontrols), full.c)
-	co.x <- X[Treatment==0,] # control group
+	co.x <- X[Treatment==0,,drop = FALSE] # control group
 	co.x <- cbind(rep(1,ncontrols),co.x)
 	
 	# treated group mean
@@ -351,7 +371,7 @@ hbal <- function(
 	if (!is.null(term.alpha)){
 		penalty.names <- names(term.alpha)
 		if (length(match(penalty.names, colnames(X)))==0) {
-			if (print.level >= 0) message("Invalid variable name(s); \"term.alpha\" is ignored\n")
+			if (print.level >= 1) message("Invalid variable name(s); \"term.alpha\" is ignored\n")
 			penalty.names <- penalty.val <- penalty.pos <- NULL
 		} else {
 			penalty.val <- term.alpha
@@ -380,14 +400,19 @@ hbal <- function(
 	if (length(coefs) != ncol(co.x)) {stop("coefs needs to have same length as number of covariates plus one")}
 
 	# Prepare for cross-validation
-	if (cv %in% c(TRUE, FALSE) == FALSE) {stop("cv needs to be of type logical")}
-	if (cv==TRUE && length(grouping)==1){
+	if (is.null(cv) == FALSE) {
+		if (cv %in% c(TRUE, FALSE) == FALSE) {stop("cv needs to be of type logical")}
+		# if (cv==TRUE && length(grouping)==1){
+		# 	cv <- FALSE
+		# 	if (print.level >= 1) {
+		# 		message("length(grouping)==1, no need to cross-validate tuning parameters. \n
+		# 			Either double selection selected 0 higher order term or the supplied grouping has length 1")}
+		# }
+	} else {
 		cv <- FALSE
-		if (print.level > 0) {
-			message("length(grouping)==1, no need to cross-validate tuning parameters. \n
-				Either double selection selected 0 higher order term or the supplied grouping has length 1")}
 	}
 	
+	
 	# check group.alpha; if correct, no need to cv
 	if (is.null(group.alpha)==FALSE) {
 		if (length(group.alpha) != length(grouping)) {
@@ -421,46 +446,64 @@ hbal <- function(
 		}
 	}	
 
+	if (linear.exact == TRUE) {
+		if (is.null(group.exact) == TRUE) {
+			group.exact <- c(1, rep(0, length(grouping)-1))
+		} else {
+			group.exact[1] <- 1
+		}
+	}
+
 
 	# Cross-validation
 	if (cv==TRUE){
+		message("Crossvalidation...\n")
 		fold.num.co <- rep(1:folds, ceiling(ncontrols/folds))
 		fold.co <- sample(fold.num.co, ncontrols, replace=FALSE) 
 		fold.num.tr <- rep(1:folds, ceiling(ntreated/folds))
 		fold.tr <- sample(fold.num.tr, ntreated, replace=FALSE)
 
-		min.c <- nloptr(x0 = rep(1, length(grouping)-1),
-                eval_f = crossValidate,
-                lb = rep(0, length(grouping)-1),
-                ub = rep(100, length(grouping)-1),
-                opts = list('algorithm'='NLOPT_LN_COBYLA',
-                            'maxeval' =200,
-			    'ftol_rel'=1e-3,
-			    'ftol_abs'=1e-5,
-			    'xtol_abs'=1e-3,
-                            'print_level'=print.level),
-                penalty.pos=penalty.pos,
-                penalty.val=penalty.val,
-                group.exact=group.exact[-1],
-                grouping=grouping,
-                folds=folds,
-                treatment = X[Treatment==1,],
-                fold.co = fold.co,
-                fold.tr=fold.tr,
-                coefs=coefs,
-                control = co.x,
-                constraint.tolerance = constraint.tolerance,
-                print.level = print.level,
-                base.weight = base.weights.co,
-                full.t=full.t,
-                full.c=full.c,
-                shuffle.treat=shuffle.treat)
+		min.c <- nloptr(x0 = rep(0, length(grouping)),
+			eval_f = crossValidate,
+			lb = rep(0, length(grouping)),
+			ub = rep(100, length(grouping)),
+			opts = list('algorithm'='NLOPT_LN_COBYLA',
+						'maxeval' =200,
+			'ftol_rel'=1e-3,
+			'ftol_abs'=1e-5,
+			'xtol_abs'=1e-3,
+			'print_level'=print.level),
+			penalty.pos=penalty.pos,
+			penalty.val=penalty.val,
+			group.exact=group.exact,
+			grouping=grouping,
+			folds=folds,
+			treatment = X[Treatment==1,],
+			fold.co = fold.co,
+			fold.tr=fold.tr,
+			coefs=coefs,
+			control = co.x,
+			constraint.tolerance = constraint.tolerance,
+			  print.level = print.level,
+			base.weight = base.weights.co,
+			full.t=full.t,
+			full.c=full.c,
+			shuffle.treat=shuffle.treat)
+
+		names(min.c)		
+		min.c$lower_bounds
+		min.c$iterations
+		min.c$message
+		min.c$solution		
+		min.c$iterations
+		min.c$objective
 
 		if(!is.null(group.exact)){
-			min.c$solution[which(group.exact[-1]==1)] <- 0
+			min.c$solution[which(group.exact==1)] <- 0
 		}
-		# save CV results: penalty (same length of covariates)
-		group.alpha <- c(0, min.c$solution)  # add 0 for the level term
+		# save CV results: penalty (same length of "grouping")
+		group.alpha <- min.c$solution		
+		
 	} #end of tuning
 
 	# expand to penalty for individual terms
@@ -485,7 +528,13 @@ hbal <- function(
 			constraint_tolerance=constraint.tolerance,
 			print_level=print.level
 			)
-
+  
+	# non converge warning
+	if(!z$converged)
+	{
+	  message("Not converged. Try setting ds = TRUE or using X.keep and exclude option.\n")
+	}
+	
 	# save weights and coefficients
 	
 
@@ -500,13 +549,18 @@ hbal <- function(
 	grouping[1] <- grouping[1]-1 
 	alpha <- alpha[-1]
 	Covar <- colnames(X.sav)
+	#rename the covar
+	for(i in 1:length(new_names)) Covar <- stringr::str_replace_all(Covar, new_names[i], str_trunc(old_names[i], 5, side="right", ellipsis=""))
+	Covar_nonum <- Covar
+	Covar <- paste(Covar, seq_along(Covar), sep = ".")
+	colnames(X.sav) <- Covar
 	names(alpha) <- Covar
 	
 	## balance table
 	bal.tab <- matrix(NA, length(Covar), 5) # tr, co, co.w, diff, diff.w
-	rownames(bal.tab) <- Covar
-	treat <- X.sav[Treatment==1,] # treated
-	control <- X.sav[Treatment==0,] # control
+	rownames(bal.tab) <- Covar_nonum
+	treat <- X.sav[Treatment==1, , drop = FALSE] # treated
+	control <- X.sav[Treatment==0, , drop = FALSE] # control
 	bal.tab[,1] <- apply(treat, 2, weighted.mean, w = base.weights.tr)
 	bal.tab[,2] <- apply(control, 2, weighted.mean, w = base.weights.co)	 
 	bal.tab[,3] <- apply(control, 2, weighted.mean, w = weights.co)	 
